use super::parse;
use anyhow::{bail, Context, Result};
use std::{fs::File, io::Write, path::Path};
use url::Url;

/// Convert relate to absolute
pub fn relative_to_absolute(base_url: &str, relative_href: &str) -> Result<String> {
    let base_url = Url::parse(base_url).context(format!("Url::parse({})", base_url))?;
    let new_url = base_url
        .join(relative_href)
        .context(format!("join({})", relative_href))?;
    Ok(new_url.to_string())
}

pub struct FetchResult {
    pub cached: bool,
    pub html: String,
}

pub fn agent_update(user_agent: &str) -> Result<String> {
    let client = reqwest::blocking::Client::builder()
        .user_agent(user_agent)
        .build()?;

    let result = fetch(&client, "https://www.mozilla.org/en-US/firefox/releases/")?;
    if let Ok(v) = parse::find_versions(&result) {
        if v != user_agent {
            return Ok(v);
        }
    }
    bail!("No user_agent updates.");
}

#[allow(dead_code)]
// Should this always fetch it/save?
pub fn fetch(client: &reqwest::blocking::Client, url: &str) -> Result<String> {
    let res = client.get(url).send()?;
    let buffer = res.text()?;
    Ok(buffer)
}

/// Extract filename from the end of a URL.
pub fn filename_from_url(url: &str) -> Result<String> {
    let (_, filename) = url.rsplit_once('/').context("Failed to split URL.")?;
    Ok(filename.to_string())
}

// Option to always fetch, ignore cache?
pub fn fetch_cache(
    work_dir: &str,
    client: &reqwest::blocking::Client,
    url: &str,
) -> Result<FetchResult> {
    let filename = filename_from_url(url)?;
    // let (_, filename) = url.rsplit_once('/').context("Failed to split URL.")?;
    let path = Path::new(work_dir).join(&filename);

    if path.exists() {
        // File already exists -- use cached version.
        let buffer = std::fs::read_to_string(&path).with_context(|| {
            format!("Failed to read file [{:?}].", &path)
            // path.to_string_lossy().to_string());
        })?;

        return Ok(FetchResult {
            cached: true,
            html: buffer,
        });
    }

    println!("fetch_cache {} => {}", url, filename);
    let res = client.get(url).send()?;
    let buffer = res.text()?;

    let mut file =
        File::create(&path).with_context(|| format!("Failed to write file [{:?}].", &path))?;

    file.write_all(buffer.as_bytes())
        .with_context(|| format!("Failed to write all to [{:?}].", &path))?;

    Ok(FetchResult {
        cached: false,
        html: buffer,
    })
}
