use anyhow::{Context, Result};
use clap::{Parser, Subcommand};
use config::save_basic_json;
use reqwest;
use scraper;
use std::fs::File;
use std::io::Write;
use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    string::String,
    sync::LazyLock,
};
use std::{thread, time::Duration};

mod config;
mod fetch;
mod parse;

const SQL_INIT: &str = include_str!("sql/init.sql");

// Setup the command line options

#[derive(Parser)]
#[command(about = "Downloads and parses Bible verses from https://www.bible.com", long_about=None, arg_required_else_help = true, after_help = "This is very specific to the website's HTML.\nIf it changes, this program might no longer work.")]
struct Cli {
    /// Working directory
    #[arg(short, long, default_value = "bible")]
    work: PathBuf,

    /// Bible Version
    #[arg(short, long, default_value = "ESV")]
    version: String,

    /// User Agent
    #[arg(short, long, action=clap::ArgAction::SetTrue)]
    agent_update: bool,

    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    /// Fetch from the web, using work directory for cache
    Fetch {
        /// Delay
        #[arg(short, long, default_value = "10")]
        delay: u32,
    },
    /// Extract information from cached files
    Extract {
        /// Count
        #[arg(short, long)] // , default_value = "5")]
        count: Option<u32>,

        /// Output file
        #[arg(short, long)]
        output: Option<PathBuf>,

        /// SQL output
        #[arg(short, long)]
        sql: Option<PathBuf>,
    },
    /// Verse of the day
    Verse {
        /// Fetch new version
        #[arg(short, long, action=clap::ArgAction::SetTrue)]
        fetch: bool,
    },
    /// Test something out
    Test {},
}

/// Configuration filename
const CONFIG_FILE: &str = "app.config";

/// Verse of the Day URL
static VOD_URL: &str = "https://www.bible.com/verse-of-the-day";

static BOOKS: LazyLock<Vec<&str>> = LazyLock::new(|| {
    Vec::from([
        "GEN", "EXO", "LEV", "NUM", "DEU", "JOS", "JDG", "RUT", "1SA", "2SA", "1KI", "2KI", "1CH",
        "2CH", "EZR", "NEH", "EST", "JOB", "PSA", "PRO", "ECC", "SNG", "ISA", "JER", "LAM", "EZK",
        "DAN", "HOS", "JOL", "AMO", "OBA", "JON", "MIC", "NAM", "HAB", "ZEP", "HAG", "ZEC", "MAL",
        "MAT", "MRK", "LUK", "JHN", "ACT", "ROM", "1CO", "2CO", "GAL", "EPH", "PHP", "COL", "1TH",
        "2TH", "1TI", "2TI", "TIT", "PHM", "HEB", "JAS", "1PE", "2PE", "1JN", "2JN", "3JN", "JUD",
        "REV",
    ])
});

#[allow(unused)]
static BOOK_NAMES: LazyLock<Vec<&str>> = LazyLock::new(|| {
    Vec::from([
        "Genesis",
        "Exodus",
        "Leviticus",
        "Numbers",
        "Deuteronomy",
        "Joshua",
        "Judges",
        "Ruth",
        "1 Samuel",
        "2 Samuel",
        "1 Kings",
        "2 Kings",
        "1 Chronicles",
        "2 Chronicles",
        "Ezra",
        "Nehemiah",
        "Esther",
        "Job",
        "Psalm",
        "Proverbs",
        "Ecclesiastes",
        "Song of Solomon",
        "Isaiah",
        "Jeremiah",
        "Lamentations",
        "Ezekiel",
        "Daniel",
        "Hosea",
        "Joel",
        "Amos",
        "Obadiah",
        "Jonah",
        "Micah",
        "Nahum",
        "Habakkuk",
        "Zephaniah",
        "Haggai",
        "Zechariah",
        "Malachi",
        "Matthew",
        "Mark",
        "Luke",
        "John",
        "Acts",
        "Romans",
        "1 Corinthians",
        "2 Corinthians",
        "Galatians",
        "Ephesians",
        "Philippians",
        "Colossians",
        "1 Thessalonians",
        "2 Thessalonians",
        "1 Timothy",
        "2 Timothy",
        "Titus",
        "Philemon",
        "Hebrews",
        "James",
        "1 Peter",
        "2 Peter",
        "1 John",
        "2 John",
        "3 John",
        "Jude",
        "Revelation",
    ])
});

static BOOK_MAP: LazyLock<HashMap<&str, usize>> =
    LazyLock::new(|| HashMap::from_iter(BOOKS.iter().enumerate().map(|x| (*x.1, x.0 + 1))));

/// find_files in base_dir that end with extension bible version.
fn find_files(base_dir: &str, version: &str) -> Vec<String> {
    let paths = std::fs::read_dir(base_dir).unwrap();
    let mut result = Vec::<String>::new();
    let version_ext = ".".to_string() + version;

    for path in paths {
        if let Ok(dir) = path {
            let filename = dir.file_name().to_string_lossy().to_string();
            if filename.ends_with(&version_ext) {
                result.push(filename);
            }
        }
    }

    // Sort filenames by the order of the books in BOOKS.

    let sorter_helper = |x: &String| -> (usize, i32) {
        let v: Vec<&str> = x.split(".").collect();
        let mut b: usize = 0;
        if BOOK_MAP.contains_key(v[0]) {
            b = BOOK_MAP[v[0]];
        }
        let c: i32 = v[1].parse().unwrap_or(0);
        (b, c)
    };

    // 1. Make it work. 2. Make it fast.

    // It would be nice to sort these (by book and chapter), so they are in order.
    // Should I just return file_names instead of path?
    result.sort_by(|a, b| {
        let a_v = sorter_helper(a);
        let b_v = sorter_helper(b);
        a_v.cmp(&b_v)
    });
    result
}

// And maybe:
// https://www.bible.com/bible/2692/GEN.1.NASB2020
// https://www.bible.com/bible/1/GEN.1.KJV
// https://www.bible.com/bible/2692/GEN.1.NASB2020
// https://www.bible.com/bible/111/GEN.1.NIV
// https://www.bible.com/bible/821/GEN.1.YLT98
// Catholic
// https://www.bible.com/bible/42/GEN.1.CPDV

// Audio
// https://www.bible.com/audio-bible/59/GEN.1.ESV
// <script type="application/ld+json">{"@context":"https://schema.org","@type":"AudioObject","mainEntityOfPage":{"@type":"WebPage","@id":"https://www.bible.com/audio-bible/59/GEN.1.ESV"},"headline":"Audio Bible: Listen to Genesis 1 English Standard Version 2016 ESV","contentUrl":"https://audio-bible-cdn.youversionapi.com/1/32k/GEN/1-9dcefc68c6f7244489f59475fc7a1866.mp3?version_id=59",

// https://www.bible.com/verse-of-the-day

fn main() -> Result<()> {
    let mut config = config::read_config(CONFIG_FILE)?;

    let cli = Cli::parse();
    // println!("Work Dir: {:?}", cli.work);
    // println!("Bible: {:?}", cli.bible);

    if !config.versions.contains_key(cli.version.as_str()) {
        println!("Sorry, I don't know about Bible Version [{}].", cli.version);
        println!("I do know about the following:");

        // Keys in arbitrary order.
        for (name, _) in config.versions.iter() {
            println!("  {}", name);
        }
        return Ok(());
    }

    if cli.agent_update {
        if let Ok(new_agent) = fetch::agent_update(&config.user_agent) {
            config.user_agent = new_agent;
            println!("User agent now {}", &config.user_agent);
            config::write_config(CONFIG_FILE, &config)?;
        } else {
            println!("User agent OK.");
        }
    }

    match &cli.command {
        Some(Commands::Fetch { delay }) => {
            let client = reqwest::blocking::Client::builder()
                .user_agent(&config.user_agent)
                .build()?;
            // .unwrap();
            let mut url = config.versions[cli.version.as_str()].to_string();
            println!("Fetch! [{}] with delay {} secs {}", cli.version, delay, url);
            let mut more = true;
            let mut cache_hit_once = true;

            while more {
                let result = fetch::fetch_cache(
                    cli.work
                        .as_os_str()
                        .to_str()
                        .expect("Work should be valid."),
                    &client,
                    url.as_str(),
                )
                .context(format!("Failed fetch {}", url))?;

                let next_chapter = parse::find_next_chapter(&result.html);

                if let Ok(next_url) = next_chapter {
                    // println!("next_url: {}", next_url);
                    let abs_url = fetch::relative_to_absolute(&url, &next_url)?;
                    // println!("now: {}", abs_url);
                    url = abs_url;
                } else {
                    // We didn't find the Next Chapter link, so stop.
                    println!(
                        "Did not find Next Chapter link. {}",
                        next_chapter.unwrap_err()
                    );
                    more = false;
                }

                // If there's more to do, add a delay between requests.
                if more {
                    if !result.cached {
                        thread::sleep(Duration::from_secs(*delay as u64));
                    } else {
                        if cache_hit_once {
                            // Display this message only once.
                            println!("Using CACHE.");
                            cache_hit_once = false;
                        }
                    }
                }
            }
            println!("I'm finished fetching!");
        }

        Some(Commands::Extract { count, output, sql }) => {
            println!("Extract...");
            let files = find_files(cli.work.to_str().unwrap(), cli.version.as_str());
            let filepath = Path::new(&cli.work);

            // let mut chapters: HashMap<String, String> = HashMap::<String, String>::new();
            let mut json_output = config::BasicJSON {
                books: Vec::new(),
                book: HashMap::new(),
            };

            let mut last_book = String::new();

            let mut extractor = |file| -> Result<()> {
                println!("File: {}", file);
                let bv = parse::extract_verses(filepath.join(file).to_str().unwrap())?;

                println!("Book {} Chapter {} BV: {:?}", bv.0, bv.1, bv.2);
                if bv.0 != last_book {
                    last_book = bv.0.clone();
                    json_output.add_to_books(&bv.0);
                }

                let json_book = json_output.new_book(&bv.0);
                let chapter = json_book.new_chapter(bv.1 as usize);

                for (idx, bv_item) in bv.2.verses.into_iter().enumerate() {
                    let verse = chapter.verse(idx + 1);
                    for bvi in bv_item {
                        verse.push(bvi);
                    }
                }
                Ok(())
            };

            if let Some(count) = *count {
                // Ok, they gave us a value.  Use it.
                println!("Extract {}:", count);
                for file in files.iter().take(count as usize) {
                    extractor(file)?;
                }
            } else {
                println!("Extract All:");
                for file in files.iter() {
                    extractor(file)?;
                }
            }

            if let Some(output) = output {
                // Ok, they gave us a file to use.
                println!("Saving output: {}", output.to_str().unwrap());
                save_basic_json(output.to_str().unwrap(), &json_output)?;
            }

            if let Some(sql_file) = sql {
                // SQL output
                println!("Saving SQL: {}", sql_file.to_str().unwrap());
                let mut file = File::create(sql_file)?;

                let _ = writeln!(file, "{}", SQL_INIT);

                // Database tables are ready.
                let mut book_id = HashMap::<String, usize>::new();

                let base = "INSERT INTO Books(book) VALUES ";
                let mut collection = Vec::<String>::new();

                for (bidx, book) in (&json_output.books).into_iter().enumerate() {
                    collection.push(format!("(\"{}\")", book));
                    // Save index number
                    book_id.insert(book.clone(), bidx + 1);
                    // writeln!(file, "INSERT INTO Books(book) VALUES (\"{}\");", book);
                }
                let _ = writeln!(file, "BEGIN TRANSACTION;");
                let _ = writeln!(file, "{} {};", base, collection.join(","));
                let _ = writeln!(file, "COMMIT;");

                // This outputs the books in a random order (according to how
                // the String keys are arranged in the HashMap).
                // I would like them to be in order.

                for book in json_output.books {
                // for (book, info) in json_output.book {
                    let info = json_output.book.get(&book).unwrap();
                    // Find the index # we need.

                    let mut bid: usize = 0;
                    let mut bid_found = false;

                    if book_id.contains_key(&book) {
                        bid = *book_id.get(&book).unwrap();
                        bid_found = true;
                    }
                    /*
                    for (idx,b) in BOOK_NAMES.iter().enumerate() {
                        if book == *b  {
                            bid = idx;
                            bid_found = true;
                            break;
                        }
                    }
                    */

                    if !bid_found {
                        panic!("I couldn't find index for {}!", book);
                    }

                    let base =
                        "INSERT INTO Verses(bid, chapter, verse, sequence, type, text) VALUES ";
                    collection.clear();

                    for (cidx, ch) in info.chapters.iter().enumerate() {
                        // Chapter would be cidx + 1
                        let _ = writeln!(file, "BEGIN TRANSACTION;");
                        for (vidx, v) in ch.verses.iter().enumerate() {
                            // Verse would be vidx + 1

                            for (pidx, p) in v.iter().enumerate() {
                                let mut t: String; //  = String::new();
                                let text: String; // = String::new();

                                match p {
                                    config::BasicVerseJSON::Heading(txt) => {
                                        t = String::from("H");
                                        text = txt.clone();
                                    }
                                    config::BasicVerseJSON::Note(txt) => {
                                        t = String::from("N");
                                        text = txt.clone();
                                    }
                                    config::BasicVerseJSON::Verse {
                                        text: txt,
                                        paragraph: p,
                                        quote: q,
                                        red: r,
                                    } => {
                                        t = String::from("V");
                                        if *p {
                                            t.push_str("P");
                                        }
                                        if *q {
                                            t.push_str("Q");
                                        }
                                        if *r {
                                            t.push_str("R");
                                        }
                                        text = txt.clone();
                                    }
                                }
                                // Part would be pidx + 1
                                collection.push(format!(
                                    "({},{},{},{},\"{}\",\"{}\")",
                                    bid,
                                    cidx + 1,
                                    vidx + 1,
                                    pidx + 1,
                                    t,
                                    text
                                ));
                            }
                        }
                        // Output the insert statement
                        if !collection.is_empty() {
                            let _ = writeln!(file, "{} {};", base, collection.join(","));
                            collection.clear();
                        }

                        let _ = writeln!(file, "COMMIT;");
                        collection.clear();
                    }
                }
            }
        }

        Some(Commands::Verse { fetch: _ }) => {
            let client = reqwest::blocking::Client::builder()
                .user_agent(&config.user_agent)
                .build()?;

            println!("Verse of the day.");
            let result = fetch::fetch_cache(
                cli.work
                    .as_os_str()
                    .to_str()
                    .expect("Work should be valid."),
                &client,
                VOD_URL,
            )?;
            if result.cached {
                println!("(from cache):");
            }
            for v in parse::find_vod(&result.html)? {
                println!("Date:  {}", v.date);
                println!("Verse: {}", v.verse);
                println!("Ref:   {}", v.reference);
                println!("------");
            }
        }

        Some(Commands::Test {}) => {
            println!("Testing...");

            if true {
                let path = Path::new(&cli.work).join("JHN.3.ESV");
                println!("Extract Verses: {:?}", path);
                parse::extract_verses(path.to_str().unwrap())?;
            }

            if false {
                let client = reqwest::blocking::Client::builder()
                    .user_agent(&config.user_agent)
                    .build()?;

                // They are using react.  There's a token request, which allows them to fetch the daily reading...
                let odb = fetch::fetch(&client, "https://www.odbm.org/");
                // See the .har file for more details.

                if let Ok(html) = odb {
                    println!("{}", html);
                } else {
                    println!("Fetch error: {:?}", odb.unwrap_err());
                }
            }

            if false {
                let path = Path::new(&cli.work).join("GEN.1.NIV");
                let buffer = std::fs::read_to_string(path).unwrap();
                let document = scraper::Html::parse_document(&buffer);

                let span_data_usfm = scraper::Selector::parse("span[data-usfm]").unwrap();
                let _span_class = scraper::Selector::parse("span[class]").unwrap();
                let span_selector = scraper::Selector::parse("span").unwrap();

                for span in document.select(&span_data_usfm) {
                    if let Some(data) = span.attr("data-usfm") {
                        println!("data-usfm {}:", data);
                        let mut lines = Vec::<&str>::new();
                        for data_span in span.select(&span_selector) {
                            if let Some(data_class) = data_span.attr("class") {
                                if data_class.contains("content") {
                                    let mut text = data_span.text().collect::<Vec<_>>();
                                    println!("{} {:?}", data, text);
                                    lines.append(&mut text);
                                }
                            }
                        }
                        println!("data {} lines {:?}", data, lines);
                    }
                }
            }
            /*
            // Test finding div>a[href^="/bible/"]
            // let a_selector = scraper::Selector::parse("div>a").unwrap();
            let a_selector = scraper::Selector::parse(r#"div>a[href ^="/bible/"]"#).unwrap();

            // This reduces down the number of items to check (from ~40 to 4)!
            // And invalidates the check for /bible/ !

            for a in document.select(&a_selector) {
                let text = a.text().collect::<Vec<_>>();
                println!("text: {:?}", text);
                if let Some(href) = a.attr("href") {
                    println!("href = {}", href);
                }
                println!("=====");
            }
            */
        }
        None => {
            println!("I didn't see a command.  Displaying help.\n");
            let _show_help: Cli = Cli::parse_from(["--help"]);
        }
    }

    Ok(())
}
